# ruff: noqa: E501
from datetime import datetime
from typing import Literal, Optional, Union

from pydantic import BaseModel, Field, HttpUrl
from pydantic.types import NonNegativeInt

from .survey import SurveyResponse


class TwitterEngagements(BaseModel):
    """Engagement counts from Twitter"""

    retweet: NonNegativeInt
    like: NonNegativeInt
    comment: NonNegativeInt
    share: NonNegativeInt


class RedditEngagements(BaseModel):
    """Engagement counts from Reddit"""

    upvote: Optional[NonNegativeInt] = Field(
        description="The reddit upvote field is deprecated, use score instead",
        deprecated=True,
        default=None,
    )
    downvote: Optional[NonNegativeInt] = Field(
        description="The reddit downvote field is deprecated, use score instead",
        deprecated=True,
        default=None,
    )
    score: Optional[int] = Field(
        description="The reddit post score (sum of upvote/downvote)", default=None
    )
    comment: NonNegativeInt
    award: NonNegativeInt


class FacebookEngagements(BaseModel):
    """Engagement counts from Facebook"""

    like: NonNegativeInt
    love: NonNegativeInt
    care: NonNegativeInt
    haha: NonNegativeInt
    wow: NonNegativeInt
    sad: NonNegativeInt
    angry: NonNegativeInt
    comment: NonNegativeInt
    share: NonNegativeInt


class ContentItem(BaseModel):
    """A content item to be ranked"""

    id: str = Field(
        description="A unique ID describing a specific piece of content. We will do our best to make an ID for a given item persist between requests, but that property is not guaranteed."
    )

    original_rank: Optional[NonNegativeInt] = Field(
        description="The rank of the item in the original feed. Useful for debugging and analysis of performance.",
        default=None,
    )

    post_id: Optional[str] = Field(
        description="The ID of the post to which this comment belongs. Useful for linking comments to their post when comments are shown in a feed. Currently this UX only exists on Facebook.",
        default=None,
    )

    parent_id: Optional[str] = Field(
        description="For threaded comments, this identifies the comment to which this one is a reply. Blank for top-level comments.",
        default=None,
    )

    title: Optional[str] = Field(
        description="The post title, only available on reddit posts.", default=None
    )

    text: str = Field(
        description="The text of the content item. Assume UTF-8, and that leading and trailing whitespace have been trimmed."
    )

    author_name_hash: str = Field(
        description="A hash of the author's name (salted). Use this to determine which posts are by the same author. When the post is by the current user, this should match `session.user_name_hash`."
    )

    type: Literal["post", "comment"] = Field(
        description="Whether the content item is a `post` or `comment`. On Twitter, tweets will be identified as `comment` when they are replies displayed on the page for a single tweet."
    )

    embedded_urls: Optional[list[HttpUrl]] = Field(
        description="A list of URLs that are embedded in the content item. This could be links to images, videos, or other content. They may or may not also appear in the text of the item."
    )

    created_at: datetime = Field(
        description="The time that the item was created in UTC, in `YYYY-MM-DD hh:mm:ss` format, at the highest resolution available (which may be as low as the hour)."
    )

    engagements: Union[TwitterEngagements, RedditEngagements, FacebookEngagements] = Field(
        description="Engagement counts for the content item."
    )


class Session(BaseModel):
    """Data that is scoped to the user's browsing session (generally a single page view)"""

    session_id: str = Field(
        description="A unique ID for this page view, updated on navigation events. Use this to determine if two requests came from the same page."
    )
    user_id: str = Field(
        description="A unique id for this study participant. Will remain fixed for the duration of the experiment."
    )
    user_name_hash: str = Field(
        description="A (salted) hash of the user's username. We'll do our best to make it match the `item.author_name_hash` on posts authored by the current user."
    )
    cohort: Optional[str] = Field(
        description="The cohort to which the user has been assigned. You can most likely ignore this. It is used by the PRC request router."
    )
    cohort_index: Optional[NonNegativeInt] = Field(
        description="The user's randomly-assigned cohort index. You can ignore this. The request router uses it to place users into buckets (cohorts).",
        default=None,
    )
    platform: Literal["twitter", "reddit", "facebook"] = Field(
        description="The platform on which the user is viewing content."
    )
    url: HttpUrl = Field(
        description="The URL of the page that the user is viewing, minus the query string portion. This can help you to determine which part of the application the user is in."
    )
    current_time: datetime = Field(
        description="The current time according to the user's browser, in UTC, in `YYYY-MM-DD hh:mm:ss` format."
    )


class RankingRequest(BaseModel):
    """A complete ranking request"""

    session: Session = Field(description="Data that is scoped to the user's browsing session")
    survey: Optional[SurveyResponse] = Field(
        description="Responses to PRC survey. Added by the request router.",
        default=None,
    )
    items: list[ContentItem] = Field(description="The content items to be ranked.")



from ranking_challenge import fake
from ranking_challenge.request import RankingRequest


def test_fake_request():
    # this test's purpose is mostly to run the code to make sure it doesn't
    # have any validation errors. pydantic will make sure it has the right fields.
    request = fake.fake_request(n_posts=5)
    assert len(request.items) == 5

    # all ids are unique
    assert len(set(item.id for item in request.items)) == 5

    request = fake.fake_request(n_posts=5, n_comments=2, platform="twitter")
    assert len(request.items) == 15
    assert request.session.platform == "twitter"


def test_fake_response():
    ids = [str(i) for i in range(5)]

    response = fake.fake_response(ids, 2)
    assert len(response.ranked_ids) == 7

    # all ids are unique
    assert len(set(id for id in response.ranked_ids)) == 7


def test_load_fake_data():
    # This really just exercises pydantic, and is mostly an example
    # of how to load json data
    request = fake.fake_request(5)
    json_data = request.model_dump_json()

    loaded_request = RankingRequest.model_validate_json(json_data)
    assert len(loaded_request.items) == 5



# ruff: noqa: E501
from enum import IntEnum
from typing import Annotated, Literal, Optional

from pydantic import BaseModel, Field


class IdeologyEnum(IntEnum):
    """Enum for ideology"""

    extremely_liberal = 1
    liberal = 2
    slightly_liberal = 3
    moderate = 4
    slightly_conservative = 5
    conservative = 6
    extremely_conservative = 7


class AgeEnum(IntEnum):
    """Enum for age brackets"""

    eighteen_to_twenty_four = 1
    twenty_five_to_thirty_four = 2
    thirty_five_to_forty_four = 3
    forty_five_to_fifty_four = 4
    fifty_five_to_sixty_four = 5
    sixty_five_and_older = 6


class EducationEnum(IntEnum):
    """Enum for education levels"""

    some_high_school = 1
    graduated_high_school = 2
    some_college = 3
    associate_degree = 4
    bachelor_degree = 5
    graduate_degree = 6


class IncomeEnum(IntEnum):
    """Enum for income brackets"""

    less_than_20k = 1
    twenty_to_thirty_five_k = 2
    thirty_five_to_fifty_k = 3
    fifty_to_seventy_five_k = 4
    seventy_five_to_one_hundred_k = 5
    one_hundred_to_one_hundred_fifty_k = 6
    one_hundred_fifty_k_and_up = 7


class SocmedUseEnum(IntEnum):
    """Enum for social media usage"""

    zero_to_thirty_minutes = 1
    thirty_to_sixty_minutes = 2
    sixty_to_ninety_minutes = 3
    ninety_to_one_twenty_minutes = 4
    two_to_three_hours = 5
    three_to_four_hours = 6
    more_than_four_hours = 7


class SocmedLeanEnum(IntEnum):
    """Enum for social media feed lean"""

    strong_liberal = 1
    moderate_liberal = 2
    neutral = 3
    moderate_conservative = 4
    strong_conservative = 5


class SurveyResponse(BaseModel):
    """
    Response to PRC survey.

    Scalar quantities are represented as IntEnums, while categorical questions are represented as strings constrained with Literal.

    These will be added to the request by the PRC request router, since this data is not available to the browser extension.
    """

    # Demographic metrics
    party_id: Literal["democrat", "republican", "independent", "other", "no_preference"] = Field(
        description="Generally speaking, do you usually think of yourself as a Republican, Democrat, Independent, etc?"
    )

    party_write_in: Optional[str] = Field(
        description="If you selected 'other' for your party identification, please specify.",
        default=None,
    )

    support: Optional[Literal["strong", "not_strong"]] = Field(
        description="Would you call yourself a strong or not a very strong supporter of your party?",
        default=None,
    )

    party_lean: Optional[Literal["democrat", "republican"]] = Field(
        description="Do you think of yourself as closer to the Republican or Democratic party?",
        default=None,
    )

    sex: Optional[Literal["female", "male", "nonbinary", "prefer_not_to_say"]] = Field(
        description="Participant's gender", default=None
    )

    age: Optional[AgeEnum] = Field(description="What age are you?", default=None)

    education: Optional[EducationEnum] = Field(
        description="What is the highest level of education you have completed?",
        default=None,
    )

    ideology: Optional[IdeologyEnum] = Field(
        description="Here is a scale on which the political views that people might hold are arranged from liberal to conservative. Where would you place yourself on this scale?",
        default=None,
    )

    income: Optional[IncomeEnum] = Field(
        description="What is your annual household income?", default=None
    )

    ethnicity: Optional[
        Literal[
            "native_american",
            "asian_or_pacific_islander",
            "black_or_african_american",
            "hispanic_or_latino",
            "white_or_caucasian",
            "multiple_or_other",
        ]
    ] = Field(description="Which race or ethnicity best describes you?", default=None)

    ethnicity_write_in: Optional[str] = Field(
        description="If you selected 'multiple' or 'other' for your ethnicity, please specify.",
        default=None,
    )

    socmed_use: Optional[SocmedUseEnum] = Field(
        description="Think of the past two weeks. How much time did you spend on social media, on average, per day?",
        default=None,
    )

    feed_lean: Optional[SocmedLeanEnum] = Field(
        description="Do you think your social media feeds tend to favour liberal content, conservative content, or neither?",
        default=None,
    )

    socmed_censorship: Optional[
        Literal["not_at_all_likely", "not_very_likely", "somewhat_likely", "very_likely"]
    ] = Field(
        description="How likely, if at all, do you think it is that social media sites intentionally censor political viewpoints that they find objectionable?",
        default=None,
    )

    socmed_trust: Optional[
        Literal[
            "strongly_distrust",
            "somewhat_distrust",
            "neither_trust_nor_distrust",
            "somewhat_trust",
            "strongly_trust",
        ]
    ] = Field(
        description="On balance, to what extent do you trust that you receive accurate information from social media?",
        default=None,
    )

    percieved_racism: Optional[
        Literal[
            "not_a_problem",
            "slight_problem",
            "moderate_problem",
            "significant_problem",
            "major_problem",
        ]
    ] = Field(
        description="To what extent do you view racism against Black Americans as a problem today?",
        default=None,
    )

    trump: Optional[
        Literal[
            "strongly_unfavorable",
            "somewhat_unfavorable",
            "neither_favorable_nor_unfavorable",
            "somewhat_favorable",
            "strongly_favorable",
        ]
    ] = Field(
        description="How favorable/unfavorable are your views towards Donald Trump?",
        default=None,
    )

    economic: Optional[
        Literal[
            "extremely_negative",
            "somewhat_negative",
            "neither_positive_nor_negative",
            "somewhat_positive",
            "extremely_positive",
        ]
    ] = Field(
        description="How positive/negative do you feel about the current U.S. economic situation?",
        default=None,
    )

    msm_trust: Optional[
        Literal[
            "strongly_distrust",
            "somewhat_distrust",
            "neither_trust_nor_distrust",
            "somewhat_trust",
            "strongly_trust",
        ]
    ] = Field(
        description="How much do you trust mainstream mass media--such as newspapers, TV and radio--to report the news fully, accurately, and fairly?",
        default=None,
    )

    immigration: Optional[
        Literal[
            "greatly_decreased",
            "somewhat_decreased",
            "kept_the_same",
            "somewhat_increased",
            "greatly_increased",
        ]
    ] = Field(
        description="In your view, should immigration to the U.S. be increased, decreased, or kept about the same?",
        default=None,
    )

    israel_palestine: Optional[
        Literal[
            "strongly_oppose",
            "somewhat_oppose",
            "neither_support_nor_oppose",
            "somewhat_support",
            "strongly_support",
        ]
    ] = Field(
        description="How strongly do you support/oppose Israel in the Israeli-Palestinian conflict?",
        default=None,
    )

    abortion: Optional[
        Literal[
            "strongly_oppose",
            "somewhat_oppose",
            "neither_support_nor_oppose",
            "somewhat_support",
            "strongly_support",
        ]
    ] = Field(
        description="How strongly do you support/oppose the legal right to abortion?",
        default=None,
    )

    climate_change: Optional[
        Literal[
            "not_concerned",
            "slightly_concerned",
            "moderately_concerned",
            "very_concerned",
            "extremely_concerned",
        ]
    ] = Field(
        description="How concerned are you about climate change?",
        default=None,
    )

    military: Optional[
        Literal[
            "strongly_unfavorable",
            "somewhat_unfavorable",
            "neither_favorable_nor_unfavorable",
            "somewhat_favorable",
            "strongly_favorable",
        ]
    ] = Field(
        description="How favorable/unfavorable are your views towards the U.S. military?",
        default=None,
    )

    political_complexity: Optional[
        Literal["never", "some_of_the_time", "half_the_time", "most_of_the_time", "always"]
    ] = Field(
        description="How often do politics and government seem so complicated that you can't really understand what's going on?",
        default=None,
    )

    political_understanding: Optional[
        Literal[
            "extremely_well",
            "very_well",
            "moderately_well",
            "slightly_well",
            "not_well_at_all",
        ]
    ] = Field(
        description="How well do you understand the important political issues facing our country?",
        default=None,
    )

    political_focus: Optional[
        Literal["never", "some_of_the_time", "half_the_time", "most_of_the_time", "always"]
    ] = Field(
        description="How often do you pay attention to what’s going on in government and politics?",
        default=None,
    )

    voting_likelihood: Optional[
        Literal[
            "will_not_vote",
            "probably_will_not_vote",
            "probably_will_vote",
            "definitely_will_vote",
        ]
    ] = Field(
        description="How likely are you to vote in the general election this November?",
        default=None,
    )

    browser_perc_twitter: Annotated[
        float | None,
        Field(
            ge=0,
            le=1,
            default=None,
            description="For Twitter, how often would you say you use social media from your desktop computer (including laptops), as opposed to your mobile?",
        ),
    ]

    browser_perc_facebook: Annotated[
        float | None,
        Field(
            ge=0,
            le=1,
            default=None,
            description="For Facebook, how often would you say you use social media from your desktop computer (including laptops), as opposed to your mobile?",
        ),
    ]

    browser_perc_reddit: Annotated[
        float | None,
        Field(
            ge=0,
            le=1,
            default=None,
            description="For Reddit, how often would you say you use social media from your desktop computer (including laptops), as opposed to your mobile?",
        ),
    ]



from typing import Optional
from uuid import uuid4

from pydantic import BaseModel, Field, HttpUrl


class NewItem(BaseModel):
    """A new item to be inserted into the feed"""

    id: str = Field(
        description="A unique ID for the content item. You can generate this.",
        default_factory=uuid4,
    )
    url: HttpUrl = Field(description="The publicly-accessible URL of the content item.")


class RankingResponse(BaseModel):
    """A response to a ranking request"""

    ranked_ids: list[str] = Field(
        description="The IDs of the content items, in the order they should be displayed."
    )
    new_items: Optional[list[NewItem]] = Field(
        description="New publicly-accessible items to be inserted into the feed.",
        default=None,
    )



import logging
import os
import json

from sqlalchemy.engine.url import URL
from urllib.parse import urlparse, parse_qs

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s.%(msecs)03d [%(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S,%f",
)

logger = logging.getLogger(__name__)


def build_db_uri(source_uri=None):
    """This is a helper function to inject postgres credentials from env vars
    into a connection string.

    Args:
        source_uri (Optional[str]): base connection string, if not provided, will
            use the DB_URI environment variable

    It is needed because the URI components may be passed in the DB_URI environment
    variable in their entirety (e.g. in local testing), but in production, the secrets
    secrets are generally passed in as separate environment variables.

    This function implements the following logic:
    - if PGPASSWORD is set, use it as the password and PGUSER (default: postgres)
      as the username
    - if POSTGRES_CREDENTIALS is set, assume it contains a JSON object of the form
          {"username": "myuser", "password": "mypass"}
    - if none of the above conditions apply, return DB_URI as given
    """
    if source_uri is None:
        source_uri = os.getenv("DB_URI")
    if source_uri is None:
        logger.error("DB_URI not set and source_uri not provided")
        raise Exception("DB_URI not set and source_uri not provided")
    parsed_uri = urlparse(source_uri)
    if parsed_uri.scheme == "":
        logger.error("Invalid URI scheme")
        raise Exception("Invalid URI scheme")
    query = {}
    if parsed_uri.query:
        query = parse_qs(parsed_uri.query)
    username = parsed_uri.username or "postgres"
    password = ""
    if os.getenv("PGPASSWORD"):
        password = os.getenv("PGPASSWORD")
    if os.getenv("PGUSER"):
        username = os.getenv("PGUSER")
    if not (username and password):
        if (env_creds := os.getenv("POSTGRES_CREDENTIALS")) is not None:
            creds = json.loads(env_creds)
            username = creds.get("username", username)
            password = creds.get("password", password)
    if not (username and password):
        return source_uri
    return URL(
        drivername=parsed_uri.scheme,
        username=username,
        password=password,
        host=parsed_uri.hostname,
        port=parsed_uri.port,
        database=parsed_uri.path.lstrip("/"),
        query=query,  # type:ignore
    )



import hashlib
import time
from random import randint
from uuid import uuid4

from faker import Faker
from ranking_challenge.request import ContentItem, RankingRequest, Session
from ranking_challenge.response import RankingResponse
from ranking_challenge.survey import SurveyResponse

fake = Faker(locale="la")  # remove locale to get rid of the fake latin

URI_PATHS = {
    "reddit": ["", "r/aww", "r/politics", "r/programming", "r/technology"],
    "twitter": ["", "jack", "TiredActor", "horse_ebooks"],
    "facebook": ["", "photo", "groups"],
}

platform = "twitter"


def fake_request(n_posts=1, n_comments=0, platform=platform):
    posts = [fake_item(platform=platform, type="post") for _ in range(n_posts)]
    comments = []
    for post in posts:
        last_comment_id = None
        for _ in range(n_comments):
            comments.append(
                fake_item(
                    platform=platform,
                    type="comment",
                    post_id=post.id,
                    parent_id=last_comment_id,
                )
            )
            last_comment_id = comments[-1].id

    return RankingRequest(
        session=Session(
            user_id=str(uuid4()),
            session_id=str(uuid4()),
            url=f"https://{platform}.com/{fake.random_element(URI_PATHS[platform])}",
            user_name_hash=hashlib.sha256(fake.name().encode()).hexdigest(),
            cohort="AB",
            cohort_index=randint(0, 4095),
            platform=platform,
            current_time=time.time(),
        ),
        survey=SurveyResponse(
            party_id="democrat",
            support="strong",
            party_lean="democrat",
            sex="female",
            age=3,
            education=4,
            ideology=5,
            income=6,
            ethnicity="native_american",
            socmed_use=7,
            browser_perc=0.8,
            mobile_perc=0.2,
            feed_lean=3,
            socmed_censorship="not_at_all_likely",
            socmed_trust="strongly_distrust",
            percieved_racism="not_a_problem",
            trump="strongly_unfavorable",
            economic="extremely_negative",
            msm_trust="strongly_distrust",
            immigration="greatly_decreased",
            israel_palestine="strongly_oppose",
            abortion="strongly_oppose",
            climate_change="not_concerned",
            military="strongly_unfavorable",
            political_complexity="never",
            political_understanding="extremely_well",
            political_focus="never",
            voting_likelihood="will_not_vote",
        ),
        items=posts + comments,
    )


def fake_item(platform="reddit", type="post", post_id=None, parent_id=None):
    if platform == "reddit":
        score = randint(-50, 50)
        if score > 0:
            upvote = score
            downvote = 0
        else:
            upvote = 0
            downvote = -score

        engagements = {
            "upvote": upvote,
            "downvote": downvote,
            "score": score,
            "comment": randint(0, 50),
            "award": randint(0, 50),
        }
    elif platform == "twitter":
        engagements = {
            "like": randint(0, 50),
            "retweet": randint(0, 50),
            "comment": randint(0, 50),
            "share": randint(0, 50),
        }
    elif platform == "facebook":
        engagements = {
            "like": randint(0, 50),
            "love": randint(0, 50),
            "care": randint(0, 50),
            "haha": randint(0, 50),
            "wow": randint(0, 50),
            "sad": randint(0, 50),
            "angry": randint(0, 50),
            "comment": randint(0, 50),
            "share": randint(0, 50),
        }
    else:
        raise ValueError(f"Unknown platform: {platform}")

    item = ContentItem(
        id=str(uuid4()),
        original_rank=randint(0, 100),
        text=fake.text(),
        post_id=post_id,
        parent_id=parent_id,
        author_name_hash=hashlib.sha256(fake.name().encode()).hexdigest(),
        type=type,
        created_at=time.time(),
        embedded_urls=[fake.url() for _ in range(randint(0, 3))],
        engagements=engagements,
    )

    return item


def fake_response(ids, n_new_items=1):
    new_items = [fake_new_item() for _ in range(n_new_items)]

    ids = list(ids) + [item["id"] for item in new_items]

    return RankingResponse(ranked_ids=ids, new_items=new_items)


def fake_new_item():
    return {
        "id": str(uuid4()),
        "url": f"https://{platform}.com/{fake.random_element(URI_PATHS[platform])}",
    }


# if run from command line
def main():
    request = fake_request(n_posts=1, n_comments=2)
    print("Request:")
    print(request.model_dump_json(indent=2))

    # use ids from request
    response = fake_response([item.id for item in request.items], 2)
    print("\nResponse:")
    print(response.model_dump_json(indent=2))


if __name__ == "__main__":
    main()
